//**************************************************************************
//This file contains individual,population representation and genetic      *
//operators.(Programmed by NXH).                                           *
//**************************************************************************
#ifndef __GENETIC__
#define __GENETIC__

#include<iostream.h>
#include<stdlib.h>
#include<string.h>
#include<assert.h>
#include<math.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "grammar.h"
//Declaration for individual and population
struct _individual {
		      node *chrom;//chromosome of individual
		      int size,height;//size and heigh of individual
		      double branch;//branching of individual
		      double fitness;//fitness of individual
		   };
typedef struct _individual individual;
individual *oldpop,*newpop;//New and old populations
unsigned int poplen; //population length
double pcross,pmutate,pcopy;//probability for genetic operations
long ncross,nmutate;//statistics
//*********************
//Copy an individual  *
//*********************
individual CopyIndividual(individual t,BOOL copychrom)
{
 individual v;
 if(copychrom)
  v.chrom=CopyNode(t.chrom);
 else
  v.chrom=t.chrom;
  v.size=t.size;
  v.height=t.height;
  v.fitness=t.fitness;
  v.branch=t.branch;
 return v;
}
//***********************
//Display an individual *
//***********************
void DisplayIndividual(individual v)
{
 cout<<endl<<"-----------------------------";
 cout<<endl<<"Genotype Structure:";
 DisplaySTree(v.chrom);
 cout<<endl<<"Size:"<<v.size;
 cout<<endl<<"Height:"<<v.height;
 cout<<endl<<"Branching:"<<v.branch;
 cout<<endl<<"Fitness:"<<v.fitness;
}
//**************************************************************
//Init the population with max depth and max try for each tree *
//**************************************************************
void RampedInit(char *startsymbol, int depth)
{
 int i,j,k;
 node *t;
 for(i=0;i<poplen;i++)
  {
   j=0;
   do
    {
      k=GetNonTerminalIndex(startsymbol);
      if(k<0)
      {
       cout<<"Invalid Start Symbol! "<<startsymbol<<endl; 
      }
      t=GenDerivationTree(k,depth);
      j++; 
    } while((j<MAXATEMPT)&& !t);
   if(!t)
   {
     cout<<endl<<"Cannot Generate Derivation trees!!!"<<endl;
     exit(1);
   }    
  // DisplaySTree(t);
   oldpop[i].chrom=t;
   oldpop[i].size=GetSizeNode(t,TRUE);
   oldpop[i].height=GetHeightNode(t);
   oldpop[i].branch=GetAVGNode(t);
 }
}
//***************************************
//Initialize the population from a file *
//***************************************
void InitFromFile(FILE *fp)
{
 int i;
 //first read the poplen;
 poplen=fscanf(fp,"%d%*c",&poplen);
 //read individual line by line
 for(i=0;i<poplen;i++)
 {
  fgets(_sbuff,MAXSTRING,fp); 
  _curchar=0;
  oldpop[i].chrom=StringToTree();
  oldpop[i].size=GetSizeNode(oldpop[i].chrom,TRUE);
  oldpop[i].height=GetHeightNode(oldpop[i].chrom);
  oldpop[i].branch=GetAVGNode(oldpop[i].chrom);
 }
}
//**********************
//Display A Population *
//********************** 
void DisplayPopulation(FILE *fp)
{
 int i;
 if(fp)
 {
 for(i=0;i<poplen;i++)
  {
   TreeToString(oldpop[i].chrom);
   fputs(_sbuff,fp);
   fprintf(fp,"\n");
   //to be added
  }
 }
else
 {
  for(i=0;i<poplen;i++)
   DisplayIndividual(oldpop[i]);
 } 
}
//************************************************
//This function computes Adjusted Fitness        *
//This assumes that the fitness of an individual *
//is already standardized (it means the smaller) *
//the better)                                    *
//************************************************
void AdjustFitness()
{
  int i;
  for(i=0;i<poplen;i++)
   oldpop[i].fitness=1/(1+oldpop[i].fitness);
}
//***********************************************
//This function computes normalized fitness it  *
//supposes that the fitness of an individual is *
//already adjusted (it means between 0 - 1 and  *
//The bigger the better)                        *
//***********************************************
void NormalizeFitness()
{
  double sum=0;
  int i;
  for(i=0;i<poplen;i++)
   sum+=oldpop[i].fitness;
  for(i=0;i<poplen;i++)
   oldpop[i].fitness=oldpop[i].fitness/sum;
}
//*************************************************
//Selections works on Normalized Fitness          *
//*************************************************
//Propotionate Selection
int ProSelect()
{
  double k=Next_Double();
  int i=0;
  while((i<poplen-1)&&(k>0))
   {
     k-=oldpop[i].fitness;
     if(k>0)
      i++;
   }
   return i;
}
//**********************
//Tournement Selection *
//**********************
int TourSelect(int size)
{
   int i,j,pos=0;
   double max=-1000;
   assert(size<=MAXTOUR);
   for(i=0;i<size;i++)
    {
      j=IRandom(0,poplen-1);
      if (oldpop[j].fitness>max)
       {
	 max=oldpop[j].fitness;
	 pos=j;
       }
    }
   return pos;
}
//Linear Ranking Selection
//**************************************************
//Crossover Section                                *
//**************************************************
BOOL SubTreeSwap(individual parent1,individual parent2,individual &child1,individual &child2)
{
 individual temp1, temp2;
 int count, crosspoint1,crosspoint2;
 node *node1,*node2,*t;
 double valuetemp;
 char nametemp[MAXNAME];
 count=0;
 while (count<MAXATEMPT)
   {
    //copy the parents
    temp1=CopyIndividual(parent1,TRUE);
    temp2=CopyIndividual(parent2,TRUE);
    //attemp to make crossover
    crosspoint1=IRandom(2,temp1.size);//excluding the root
    crosspoint2=IRandom(2,temp2.size);
    LocateNode(crosspoint1,temp1.chrom,NULL);
    node1=_idnode;
    LocateNode(crosspoint2,temp2.chrom,NULL);
    node2=_idnode;
   //check the if they are nonterminal and they are the same
   if((node1->children)&&(node2->children)&&!strcmp(node1->name,node2->name))
   { 
    //try to do crossover
    //first swaping contents
    strcpy(nametemp,node1->name);
    strcpy(node1->name,node2->name);
    strcpy(node2->name,nametemp);
    valuetemp=node1->value;
    node1->value=node2->value;
    node2->value=valuetemp;
   //then swap the children
    t=node1->children;
    node1->children=node2->children;
    node2->children=t;
   //compute the heigh after that
    temp1.height=GetHeightNode(temp1.chrom);
    temp2.height=GetHeightNode(temp2.chrom);
    if((temp1.height<=MAXDEPTH)&&(temp2.height<=MAXDEPTH))
     {
      temp1.size=GetSizeNode(temp1.chrom,TRUE);
      temp1.branch=GetAVGNode(temp1.chrom);
      temp2.size=GetSizeNode(temp2.chrom,TRUE);
      temp2.branch=GetAVGNode(temp2.chrom);
      child1=CopyIndividual(temp1,FALSE);
      child2=CopyIndividual(temp2,FALSE);
      return TRUE;
     }
    else
     {
      DeleteNode(temp1.chrom);
      DeleteNode(temp2.chrom);
     }
    }
    count++;
   }
return FALSE;
}
//**************************************************
//Mutation Section                                 *
//**************************************************
BOOL ReplaceSubTree(individual &child,int maxdepth)
{
 individual temp;
 int replacepoint,count,k;
 node *node1,*parentnode1,*t,*p;
 count=0;
while(count<MAXATEMPT)
{
 temp=CopyIndividual(child,TRUE);
 replacepoint=IRandom(1,temp.size);// the root
 LocateNode(replacepoint,temp.chrom,NULL);
 node1=_idnode;
 parentnode1=_idnodeparent;
if(replacepoint==1)//mutation at root
{
 k=GetNonTerminalIndex((temp.chrom)->name);
 t=GenDerivationTree(k,MAX(MAXDEPTH,maxdepth));
 if(t)
 {
  DeleteNode(child.chrom);
  DeleteNode(temp.chrom);
  child.chrom=t;
  child.size=GetSizeNode(t,TRUE);
  child.height=GetHeightNode(t);
  child.branch=GetAVGNode(t);
  return TRUE; 
 } 
} 
else
if(node1->children)
 { 
  k=GetNonTerminalIndex(node1->name);
  t=GenDerivationTree(k,maxdepth); 
  if(t)
  {
   p=parentnode1->children;
  if(p==node1) //first child
     parentnode1->children=t;
   else
    {
     while(p->sibling!=node1)
      p=p->sibling;
      p->sibling=t;
    }
   t->sibling=node1->sibling;
  DeleteNode(node1);
  temp.height=GetHeightNode(temp.chrom);
 if(temp.height<=MAXDEPTH)
  {
   DeleteNode(child.chrom);
   child.chrom=temp.chrom;
   child.height=temp.height;
   child.size=GetSizeNode(child.chrom,TRUE);
   temp.branch=GetAVGNode(child.chrom);
   return TRUE;
   }
  }
 }
  DeleteNode(temp.chrom);
  count++;
 }
return FALSE;
}
#endif
