#ifndef __GRAMMAR__
#define __GRAMMAR__
/* This file contains declaration and implementation of grammars */
/* GGP system @copyright  by Nguyen Xuan Hoai */
#include <iostream.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <assert.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"

#define MAXRULE 20 //maximum number of rules for one nonterminal
#define MAXRLENGTH 10 //maximum number of symbols in one rule

//*****************
//SYMBOL SECTION  *
//*****************

/* Declaration for symbol list*/
 struct  _sym {
                            char name[MAXNAME];
                            int minpath; //minimal path to go to terminal 
               };
 typedef struct _sym symbol;

//Global variable for nonterminal and terminal symbols
 symbol nonterminal [MAXNONTERMINAL];
 int nnterm;
 symbol terminal [MAXTERMINAL];
 int nterm;
 //manipulating symbols
 //*****************************************************
 //This function initialise symbol lists for a grammar *
 //*****************************************************
 void InitSymbol(void)
 {
   nnterm=0;
   nterm=0;
 }
 //***********************************************
 //Check if a tring is the name of a nonterminal *
 //***********************************************
BOOL IsNonTerminal (char  *s)
{
  int i;
  for(i=0;i<nnterm;i++)
    if(!strcmp(nonterminal[i].name,s))
	return TRUE;
 return FALSE;	
}
//************************************************************
//Get a Nonterminal index from its name return -1 if failed  *
//************************************************************
int GetNonTerminalIndex(char *s)
{
 int i;
 for(i=0;i<nnterm;i++)
    if(!strcmp(nonterminal[i].name,s))
	return i;
 return -1;
}
 //********************************************
 //Check if a tring is the name of a terminal *
 //********************************************
BOOL IsTerminal (char *s)
{
  int i;
  for(i=0;i<nterm;i++)
    if(!strcmp(terminal[i].name,s))
	return TRUE;
  return FALSE;

}
//********************************************************
//Get a terminal index from its name return -1 if failed *
//********************************************************
int GetTerminalIndex (char *s)
{
  int i;
  for(i=0;i<nterm;i++)
    if(!strcmp(terminal[i].name,s))
	return i;
  return -1;

}
//*******************************************
//Adding one symbol to the symbol lists     *
//Type=TRUE for nonterminal FALSE otherwise *
//*******************************************
 void AddSymbol(char *name,BOOL type)
 {
   if (type) //if adding non-terminal symbol
   {
         assert (nnterm<MAXNONTERMINAL);
     //check if already exist
	 if(IsNonTerminal(name)||IsTerminal(name))
	 {
	   cout<<endl<<"ERROR: Symbol duplication! "<<name<<endl;
	   exit(1);
	 }
	 strcpy(nonterminal[nnterm].name,name);
	 nonterminal[nnterm].minpath=0;
         nnterm++;
   }    
  else
   {
     assert (nterm<MAXTERMINAL);
	 if(IsTerminal(name)||IsNonTerminal(name))
	 {
	   cout<<endl<<"ERROR: Symbol duplication! "<<name<<endl;
	   exit(1);
	 }
	 strcpy(terminal[nterm].name,name);
	 terminal[nterm].minpath=0;
        nterm++;
   }
 }
//***********************
//Print the symbol list *
//***********************
void PrintSymbol(void)
{
  int i;
  cout<<endl<<"Nonterminal symbols:";
  for(i=0;i<nnterm;i++)
   cout<<endl<<"Symbol "<<i<<":"<<nonterminal[i].name<<" minpath:"<<nonterminal[i].minpath;
  cout<<endl<<"Terminal symbols:";
  for(i=0;i<nterm;i++)
   cout<<endl<<"Symbol "<<i<<":"<<terminal[i].name;
 
}

//**************
//Rule SECTION *
//**************
//Rule righthand side declaration
struct _rulerh {
                                 char name[MAXRLENGTH][MAXNAME];//list of symbol name
				 int minpath;// min path to terminating rule of the symbols
				 int nsymbol; //number of symbols
				 double prob; //probability to be chosen
                };

typedef struct _rulerh rulerh; 

struct _ruleset {
                    rulerh rhand[MAXRULE];
	            int nrhand;
                };
typedef  _ruleset ruleset;

//**********************************
//Rule righthand side manipulation *
//**********************************

//***********************************************************************
//This function check if a rule righthand side conatains terminals only *
//***********************************************************************
BOOL IsTerminalRH(rulerh rh)
{
  int i;
  for(i=0;i<rh.nsymbol;i++)
    if(IsNonTerminal(rh.name[i]))  
      return FALSE;
 return TRUE;	
}
//****************************
//Add a rule right handside  *
//**************************** 
void AddRuleRH(char *st,rulerh &rh)
{
 char temp[MAXNAME];
 int curchar,len;
 curchar=0;
while(st[curchar]!=0) //while not end
{ 
 while(st[curchar]==' ') curchar++; //skip spaces
 if(isalpha(st[curchar]))
 {
  len=0;
  while(isalnum(st[curchar])&&(len<MAXNAME-1))
  {
    temp[len]=st[curchar];
    curchar++;
   len++;
  }
  temp[len]=0;//terminate the string
  //check if it is a genuine symbol
  if(!IsNonTerminal(temp) && ! IsTerminal(temp))
  {
      cout<<endl<<"Unrecognized Symbol!!! "<<temp<<endl;
      exit(1);
  }
  //add to the right handside
  assert(rh.nsymbol<MAXRLENGTH);
  strcpy(rh.name[rh.nsymbol],temp);
  rh.nsymbol++;
  
 }
 else
 {
   cout<<"Error in String!!!";
   exit(1);
 }
} 
 
}
//******************************************************
//This fucntion to find the min depth of a rule right handside               *
//return min depth> o or 0 for terminal rule and not yet computable    *
//It is uncomputable because at that time some nonterminal has not    *
//yet labeled the mindepth (mindepth=0)                                             * 
//******************************************************
int MinDepth(rulerh rh)
{
  int i,max,j;
//looking at min depth of all nonterminal in the right handside
 max=0;
 for(i=0;i<rh.nsymbol;i++) 
 {
   j=GetNonTerminalIndex(rh.name[i]); 
   if(j>=0)//if nonterminal
    if(nonterminal[j].minpath>0)
     {
       if(max<nonterminal[j].minpath)
          max=nonterminal[j].minpath;
     }
    else
     return 0; //there are some nonterminals not labelled yet 
  }
 return max+1;
}

//global variable for ruleset of the grammar
 ruleset  *rules;
 int nrules;     

//********************************
//Initialization of the rule set *
//********************************     
void InitRuleSet(void)
{
 int i;
 nrules=nnterm;
 rules = new ruleset[nrules]; 
 assert (rules!=NULL);
 for(i=0;i<nrules;i++)
    rules[i].nrhand=0;
}
//********************
//Display a rule set *
//********************
void DisplayRule(void)
{
 int i,j,k;
 for(i=0;i<nrules;i++)
  for(j=0;j<rules[i].nrhand;j++)
   {
   //display the lefthand side
    cout<<endl<<nonterminal[i].name<<"->";
    for(k=0;k<rules[i].rhand[j].nsymbol;k++)
     cout<<rules[i].rhand[j].name[k]<<" ";
    cout<<rules[i].rhand[j].prob<<" ";
    cout<<rules[i].rhand[j].minpath; 
   }
}

//**************************
//add a rule from a string *
//**************************
void AddRule(const char *st,double weight)
{
 char temp[MAXNAME];
 int curchar,nindex,nlen,k;

 curchar=0;
 strcpy(_sbuff,st);
 while (_sbuff[curchar]==' ')
  curchar ++;
 if(isalpha(_sbuff[curchar])) //get the left handside of the rule
 {
  nlen=0;
  while(isalnum(_sbuff[curchar])&&(nlen<MAXNAME-1))
  {
   temp[nlen]=_sbuff[curchar];
   nlen++;
   curchar++;
  }
  temp[nlen]=0;
  //skip -> symbols
  while(_sbuff[curchar]==' ') curchar++;
  if(_sbuff[curchar]=='-')
   curchar+=2;
  else
  {
   cout<<endl<<"Rule Syntax Error!!!";
   exit(1);
  }	  
  //check if it is genuine nonterminal
   nindex=GetNonTerminalIndex(temp); 
   if(nindex>=0)
	{
          
          k=rules[nindex].nrhand; 
	  rules[nindex].rhand[k].nsymbol=0;
	  AddRuleRH(&_sbuff[curchar],rules[nindex].rhand[k]);
	  rules[nindex].rhand[k].prob=weight;
          rules[nindex].rhand[k].minpath=0; //default
	  rules[nindex].nrhand++;
	}
   else
	{
           cout<<endl<<"Unrecognized Nonterminal Symbol!!! "<<temp;
	   exit(1);
        }
  
 }
 else
 {
   cout<<endl<<"Error in string!!!";
   exit(1);
 }
}
//****************************************************************
//Normalise the probability distribution of righthand side rules *
//****************************************************************
//Not used yet
void NormaliseRHD(void)
{
 int i,j;
 double sum;
 for(i=0;i<nrules;i++)
  {
    sum=0;
    for(j=0;j<rules[i].nrhand;j++)
     sum+=rules[i].rhand[j].prob;    
    for(j=0;j<rules[i].nrhand;j++)
      rules[i].rhand[j].prob/=sum;
  }
}
//*******************************************************************************
//Get a right handside for one nonterminal according to probability distribution*
//Using index of lefthand side                                                  *
//*******************************************************************************
//Not used yet 
int GetRHIndex(int nindex)
{
 double t;
 int rhindex;
 t=Next_Double();
 rhindex=0;
 while((t>0)&&(rhindex<rules[nindex].nrhand))
 {
  t-=rules[nindex].rhand[rhindex].prob;
  if(t>0)
   rhindex++;  
 }
 if(rhindex<rules[nindex].nrhand)
  return rhindex;
 else
  {
   cout<<endl<<"Rule Probabilistic Distribution Error!!!";
   exit(1); 
  }
}
//*******************************************************************************
//Get a right handside for one nonterminal according to probability distribution*
//Using name of the lefthand side                                               *
//*******************************************************************************
//Not used yet
int GetRHIndex2(char *st)
{
 int i;
//find the nonterminal
 i=GetNonTerminalIndex(st);
 if(i>=0)
  return GetRHIndex(i);
 else
 {
  cout<<endl<<"There is no such nonterminal!!!"<<st;
  exit(1);
 }
}
//*********************************************
//Transform a flat rule to a tree (of depth 1)*
//********************************************* 
//Not used yet
node *RuleToTree(int nindex,int rhindex)
{
 node *st,*p,*q;
 int i;
 st=MakeNode(nonterminal[nindex].name,VOIDVALUE);
 p=MakeNode(rules[nindex].rhand[rhindex].name[0],VOIDVALUE);
 st->children=p;
 for(i=1;i<rules[nindex].rhand[rhindex].nsymbol;i++)
 {
  q=MakeNode(rules[nindex].rhand[rhindex].name[i],VOIDVALUE);
  p->sibling=q;
  p=q;
 }
 return st;
}
//*****************************************************************
//Labeling the righthand side rules with min depth from bottom up *
//*****************************************************************
void LabelRule(void)
{
 int mind,i,j;
 BOOL found;
 //label all terminal rules
 for(i=0;i<nrules;i++)
  for(j=0;j<rules[i].nrhand;j++)
   if(IsTerminalRH(rules[i].rhand[j]))
      {
       nonterminal[i].minpath=1;
       rules[i].rhand[j].minpath=1; 
      }
 //Interative labelling
found=TRUE;
while(found)
{
 found=FALSE;
 for(i=0;i<nrules;i++)
  for(j=0;j<rules[i].nrhand;j++)
    if(!IsTerminalRH(rules[i].rhand[j]))
     if(rules[i].rhand[j].minpath<=0)//only label once
     {
      mind=MinDepth(rules[i].rhand[j]);
      if(mind>0)
      {
       found=TRUE;
       rules[i].rhand[j].minpath=mind;
       if (nonterminal[i].minpath<=0)
             nonterminal[i].minpath=mind; //only label once
      } 
     }
   
 }
//Check if all rules are labeled
 for(i=0;i<nrules;i++)
  for(j=0;j<rules[i].nrhand;j++)
   if(!IsTerminalRH(rules[i].rhand[j]))
    if(rules[i].rhand[j].minpath<=0)
     {
      cout<<endl<<"Cannot Label the Rule Set Properly, check the rule set!!!";
      DisplayRule();
      exit(1);
     } 
}
//****************************************************************************
//Generate a sub-derivation tree from a nonterminal and bound by a max depth *
//****************************************************************************
node *GenDerivationTree(int nindex,int maxdepth)
{
  int i,j,k;
 //BOOL usable[MAXRULE];
 double sum,dice;
 node *p,*q,*st,*t;
 sum=0;
 //find the righthand sides that can be used to extend
 for(i=0;i<rules[nindex].nrhand;i++)
  if(rules[nindex].rhand[i].minpath<=maxdepth)
   sum+=rules[nindex].rhand[i].prob;
 if(sum>0)//if there is any right handsides 
 {
    st=MakeNode(nonterminal[nindex].name,VOIDVALUE);
  //choose the righthand side
    dice=Next_Double();
    i=0;
    while((i<rules[nindex].nrhand)&&(dice>0))
    { 
      if(rules[nindex].rhand[i].minpath<= maxdepth)
         dice-=rules[nindex].rhand[i].prob/sum;
      if(dice>0)
        i++;    
    }
  q=NULL;  
 //recursive build derivation trees 
  for(j=rules[nindex].rhand[i].nsymbol-1;j>=0;j--)
   {
    k=GetNonTerminalIndex(rules[nindex].rhand[i].name[j]);
    if(k>=0)//nonterminal
     {
       //recursive call
        p=GenDerivationTree(k,maxdepth-1);
       if(!p)// if fail
        {
          //garbage collection
          p=q;
          while(p)
           {
            t=p;
            p=p->sibling;
            DeleteNode(t);
           }
          delete st;
          return NULL;
        }
       
     }
    else//terminal
      p=MakeNode(rules[nindex].rhand[i].name[j],VOIDVALUE);

     p->sibling=q;
     q=p;
    }
  st->children=p;
  return st;  
 }
else
   return NULL;
 
}
#endif
